# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any


TRIGGER_FLOW_STATUS_CREATED = "created"
TRIGGER_FLOW_STATUS_RUNNING = "running"
TRIGGER_FLOW_STATUS_WAITING = "waiting"
TRIGGER_FLOW_STATUS_COMPLETED = "completed"
TRIGGER_FLOW_STATUS_FAILED = "failed"
TRIGGER_FLOW_STATUS_CANCELLED = "cancelled"


class TriggerFlowPauseSignal(dict[str, Any]):
    def __init__(self, interrupt: dict[str, Any]):
        super().__init__(interrupt)
        self.interrupt = interrupt



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from agently.utils import LazyImport

LazyImport.import_package("fastapi", version_constraint=">=0.104")

import json
from fastapi import Body, FastAPI, WebSocket, WebSocketDisconnect
from fastapi.encoders import jsonable_encoder
from fastapi.responses import StreamingResponse
from pydantic import BaseModel, Field, create_model

from agently.utils import FunctionShifter
from agently.types.data import SerializableMapping, SerializableValue

from typing import Any, Sequence, Callable, AsyncGenerator, Generator, Protocol, TYPE_CHECKING, ParamSpec, cast

if TYPE_CHECKING:
    from fastapi.params import Depends
    from agently.core import BaseAgent, ModelRequest, TriggerFlow, TriggerFlowExecution

P = ParamSpec("P")


class FastAPIHelperGeneratorFunction(Protocol[P]):
    def __call__(
        self,
        request_data: SerializableMapping,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Generator[SerializableValue, None, None] | AsyncGenerator[SerializableValue, None]: ...


FastAPIHelperResponseWarper = Callable[[SerializableValue | Exception], SerializableValue]


class FastAPIHelperRequestData(BaseModel):
    data: Any
    options: dict[str, Any] = Field(default_factory=dict)


class FastAPIHelper(FastAPI):
    def __init__(
        self,
        *,
        response_provider: "BaseAgent | ModelRequest | TriggerFlow | TriggerFlowExecution | FastAPIHelperGeneratorFunction",
        response_warper: FastAPIHelperResponseWarper | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.response_provider = response_provider
        self._use_default_response_warper = response_warper is None
        self.response_warper: FastAPIHelperResponseWarper = (
            response_warper if response_warper is not None else self._default_response_warper
        )
        self._typed_request_model = self._build_request_model()
        self._typed_response_model = self._build_response_model()

    def _default_response_warper(self, response_data: SerializableValue | Exception):
        if isinstance(response_data, Exception):
            status = 400
            if isinstance(response_data, ValueError):
                status = 422
            elif isinstance(response_data, TimeoutError):
                status = 504
            return {
                "status": status,
                "data": None,
                "msg": str(response_data),
                "error": self._serialize_exception(response_data),
            }
        return {
            "status": 200,
            "data": self._to_json_safe(response_data),
            "msg": None,
        }

    @staticmethod
    def _to_json_safe(value: Any) -> SerializableValue:
        return cast(SerializableValue, jsonable_encoder(value))

    def _serialize_exception(self, error: Exception) -> SerializableValue:
        error_data: dict[str, SerializableValue] = {
            "type": error.__class__.__name__,
            "module": error.__class__.__module__,
            "message": str(error),
            "args": self._to_json_safe([str(arg) for arg in getattr(error, "args", [])]),
        }

        if hasattr(error, "errors") and callable(getattr(error, "errors", None)):
            try:
                error_data["validation_errors"] = self._to_json_safe(error.errors())  # type: ignore
            except Exception:
                # Keep default shape stable even if .errors() exists but fails.
                pass

        if error.__cause__ is not None:
            error_data["cause"] = self._to_json_safe(
                {
                    "type": error.__cause__.__class__.__name__,
                    "message": str(error.__cause__),
                }
            )

        if error.__context__ is not None and not error.__suppress_context__:
            error_data["context"] = self._to_json_safe(
                {
                    "type": error.__context__.__class__.__name__,
                    "message": str(error.__context__),
                }
            )

        return error_data

    def _dump_json(self, item: SerializableValue | Exception):
        return json.dumps(
            self._to_json_safe(self.response_warper(item)),
            ensure_ascii=False,
            default=str,
        )

    def _parse_request_payload(self, payload: Any) -> FastAPIHelperRequestData:
        try:
            if isinstance(payload, BaseModel):
                payload = payload.model_dump()
            if isinstance(payload, (str, bytes)):
                if isinstance(payload, bytes):
                    payload = payload.decode("utf-8", errors="replace")
                return FastAPIHelperRequestData.model_validate_json(payload)
            return FastAPIHelperRequestData.model_validate(payload)
        except Exception as e:
            raise ValueError(
                "Invalid request payload, expected JSON object with keys 'data' and 'options'. " f"Error: { e }"
            )

    def _get_triggerflow_contract_metadata(self):
        from agently.core import TriggerFlow, TriggerFlowExecution

        if isinstance(self.response_provider, (TriggerFlow, TriggerFlowExecution)):
            return self.response_provider.get_contract_metadata()
        return {}

    def _get_triggerflow_contract_data_type(self):
        from agently.core import TriggerFlow, TriggerFlowExecution

        if isinstance(self.response_provider, (TriggerFlow, TriggerFlowExecution)):
            contract = self.response_provider.get_contract()
            return contract.initial_input
        return None

    def _get_triggerflow_result_type(self):
        from agently.core import TriggerFlow, TriggerFlowExecution

        if isinstance(self.response_provider, (TriggerFlow, TriggerFlowExecution)):
            contract = self.response_provider.get_contract()
            return contract.result
        return None

    def _build_request_model(self):
        data_type = self._get_triggerflow_contract_data_type()
        return create_model(
            f"FastAPIHelperRequestData_{ id(self) }",
            data=(data_type if data_type is not None else Any, ...),
            options=(dict[str, Any], Field(default_factory=dict)),
        )

    def _build_response_model(self):
        if not self._use_default_response_warper:
            return None
        result_type = self._get_triggerflow_result_type()
        response_data_type = result_type | None if result_type is not None else Any | None
        return create_model(
            f"FastAPIHelperResponseData_{ id(self) }",
            status=(int, ...),
            data=(response_data_type, None),
            msg=(str | None, None),
            error=(dict[str, Any] | None, None),
        )

    def _build_openapi_extra(self):
        contract_metadata = self._get_triggerflow_contract_metadata()
        if not contract_metadata:
            return None
        return {
            "x-agently-triggerflow-contract": contract_metadata,
        }

    def _make_post_handler(self):
        request_model = self._typed_request_model

        async def typed_post_handler(body: request_model = Body(default=None)):  # type: ignore[valid-type]
            return await self.post_handler(body)

        return typed_post_handler

    def _get_async_generator(
        self,
        request_data: "SerializableMapping",
        *,
        options: "SerializableMapping | None" = None,
    ) -> AsyncGenerator:
        from agently.core import BaseAgent, ModelRequest, TriggerFlow, TriggerFlowExecution

        if options is None:
            options = {}

        if isinstance(self.response_provider, BaseAgent):
            if hasattr(self.response_provider, "load_json_prompt"):
                from agently.builtins.agent_extensions import ConfigurePromptExtension

                class ConfigurePromptAgent(ConfigurePromptExtension, BaseAgent): ...

                self.response_provider = cast(ConfigurePromptAgent, self.response_provider)
                self.response_provider.load_json_prompt(json.dumps(request_data, ensure_ascii=False))
                return FunctionShifter.auto_options_func(self.response_provider.get_async_generator)(**options)
            else:
                input_data = {}
                for key, value in request_data.items():
                    if hasattr(self.response_provider, key):
                        getattr(self.response_provider, key)(value)
                    else:
                        input_data[key] = value
                self.response_provider.input(input_data)
                return FunctionShifter.auto_options_func(self.response_provider.get_async_generator)(**options)
        elif isinstance(self.response_provider, ModelRequest):
            input_data = {}
            for key, value in request_data.items():
                if hasattr(self.response_provider, key):
                    getattr(self.response_provider, key)(value)
                else:
                    input_data[key] = value
            self.response_provider.input(input_data)
            return FunctionShifter.auto_options_func(self.response_provider.get_async_generator)(**options)
        elif isinstance(self.response_provider, (TriggerFlow, TriggerFlowExecution)):
            if (
                isinstance(self.response_provider, TriggerFlowExecution)
                and "concurrency" in options
                and isinstance(options["concurrency"], int)
            ):
                self.response_provider.set_concurrency(options["concurrency"])
            return FunctionShifter.auto_options_func(self.response_provider.get_async_runtime_stream)(
                request_data, **options
            )
        elif isinstance(self.response_provider, Callable):
            generator = FunctionShifter.auto_options_func(self.response_provider)(request_data, **options)
            if isinstance(generator, Generator):
                return FunctionShifter.asyncify_sync_generator(generator)
            elif isinstance(generator, AsyncGenerator):
                return generator
            else:
                raise TypeError(
                    f"[Agently FastAPI Helper] FastAPI Helper expected response provider as Agent, ModelRequest, TriggerFlow, TriggerFlowExecution, GeneratorFunction or AsyncGeneratorFunction but got: { self.response_provider } returns { generator }"
                )
        else:
            raise TypeError(
                f"[Agently FastAPI Helper] FastAPI Helper expected response provider as Agent, ModelRequest, TriggerFlow, TriggerFlowExecution, GeneratorFunction or AsyncGeneratorFunction but got: { self.response_provider }"
            )

    async def _async_get_result(
        self,
        request_data: "SerializableMapping",
        *,
        options: "SerializableMapping | None" = None,
    ):
        from agently.core import BaseAgent, ModelRequest, TriggerFlow, TriggerFlowExecution

        if options is None:
            options = {}

        if isinstance(self.response_provider, BaseAgent):
            if hasattr(self.response_provider, "load_json_prompt"):
                from agently.builtins.agent_extensions import ConfigurePromptExtension

                class ConfigurePromptAgent(ConfigurePromptExtension, BaseAgent): ...

                self.response_provider = cast(ConfigurePromptAgent, self.response_provider)
                self.response_provider.load_json_prompt(json.dumps(request_data, ensure_ascii=False))
                return await FunctionShifter.auto_options_func(self.response_provider.async_start)(**options)
            else:
                input_data = {}
                for key, value in request_data.items():
                    if hasattr(self.response_provider, key):
                        getattr(self.response_provider, key)(value)
                    else:
                        input_data[key] = value
                self.response_provider.input(input_data)
                return await FunctionShifter.auto_options_func(self.response_provider.async_start)(**options)
        elif isinstance(self.response_provider, ModelRequest):
            input_data = {}
            for key, value in request_data.items():
                if hasattr(self.response_provider, key):
                    getattr(self.response_provider, key)(value)
                else:
                    input_data[key] = value
            self.response_provider.input(input_data)
            return await FunctionShifter.auto_options_func(self.response_provider.async_start)(**options)
        elif isinstance(self.response_provider, (TriggerFlow, TriggerFlowExecution)):
            if (
                isinstance(self.response_provider, TriggerFlowExecution)
                and "concurrency" in options
                and isinstance(options["concurrency"], int)
            ):
                self.response_provider.set_concurrency(options["concurrency"])
            return await FunctionShifter.auto_options_func(self.response_provider.async_start)(request_data, **options)
        elif isinstance(self.response_provider, Callable):
            generator = FunctionShifter.auto_options_func(self.response_provider)(request_data, **options)
            if isinstance(generator, Generator):
                return [item for item in generator]
            elif isinstance(generator, AsyncGenerator):
                return [item async for item in generator]
            else:
                raise TypeError(
                    f"[Agently FastAPI Helper] FastAPI Helper expected response provider as Agent, ModelRequest, TriggerFlow, TriggerFlowExecution, GeneratorFunction or AsyncGeneratorFunction but got: { self.response_provider } returns { generator }"
                )
        else:
            raise TypeError(
                f"[Agently FastAPI Helper] FastAPI Helper expected response provider as Agent, ModelRequest, TriggerFlow, TriggerFlowExecution, GeneratorFunction or AsyncGeneratorFunction but got: { self.response_provider }"
            )

    async def websocket_handler(self, websocket: WebSocket):
        await websocket.accept()
        try:
            while True:
                data = await websocket.receive()
                if data.get("type") == "websocket.disconnect":
                    break

                try:
                    text_data = data.get("text")
                    bytes_data = data.get("bytes")
                    request = self._parse_request_payload(text_data if text_data is not None else bytes_data)

                    async for item in self._get_async_generator(request.data, options=request.options):
                        await websocket.send_text(self._dump_json(cast("SerializableValue", item)))
                except Exception as e:
                    await websocket.send_text(self._dump_json(cast("SerializableValue | Exception", e)))
        except WebSocketDisconnect:
            return

    async def sse_handler(
        self,
        payload: str,
    ):
        async def event_stream():
            try:
                request = self._parse_request_payload(payload)
                async for item in self._get_async_generator(request.data, options=request.options):
                    yield f"data: {self._dump_json(cast('SerializableValue', item))}\n\n"
            except Exception as e:
                yield f"event: error\ndata: {self._dump_json(cast('SerializableValue | Exception', e))}\n\n"

        return StreamingResponse(
            event_stream(),
            media_type="text/event-stream",
        )

    async def post_handler(self, body: Any = Body(default=None)):
        try:
            request = self._parse_request_payload(body)
            result = await self._async_get_result(request.data, options=request.options)
            return self.response_warper(cast("SerializableValue", result))
        except Exception as e:
            return self.response_warper(cast("SerializableValue | Exception", e))

    async def get_handler(
        self,
        payload: str,
    ):
        try:
            request = self._parse_request_payload(payload)
            result = await self._async_get_result(request.data, options=request.options)
            return self.response_warper(cast("SerializableValue", result))
        except Exception as e:
            return self.response_warper(cast("SerializableValue | Exception", e))

    def use_sse(
        self,
        path: str,
        name: str | None = None,
        *,
        dependencies: "Sequence[Depends] | None" = None,
    ):
        route_kwargs = {}
        openapi_extra = self._build_openapi_extra()
        if openapi_extra is not None:
            route_kwargs["openapi_extra"] = openapi_extra
        self.get(
            path,
            name=name,
            dependencies=dependencies,
            **route_kwargs,
        )(self.sse_handler)
        return self

    def use_websocket(
        self,
        path: str,
        name: str | None = None,
        *,
        dependencies: "Sequence[Depends] | None" = None,
    ):
        self.websocket(
            path,
            name,
            dependencies=dependencies,
        )(self.websocket_handler)
        return self

    def use_post(
        self,
        path: str,
        name: str | None = None,
        *,
        dependencies: "Sequence[Depends] | None" = None,
    ):
        route_kwargs = {}
        openapi_extra = self._build_openapi_extra()
        if openapi_extra is not None:
            route_kwargs["openapi_extra"] = openapi_extra
        if self._typed_response_model is not None:
            route_kwargs["response_model"] = self._typed_response_model
            route_kwargs["response_model_exclude_unset"] = True
        self.post(
            path,
            name=name,
            dependencies=dependencies,
            **route_kwargs,
        )(self._make_post_handler())
        return self

    def use_get(
        self,
        path: str,
        name: str | None = None,
        *,
        dependencies: "Sequence[Depends] | None" = None,
    ):
        route_kwargs = {}
        openapi_extra = self._build_openapi_extra()
        if openapi_extra is not None:
            route_kwargs["openapi_extra"] = openapi_extra
        if self._typed_response_model is not None:
            route_kwargs["response_model"] = self._typed_response_model
            route_kwargs["response_model_exclude_unset"] = True
        self.get(
            path,
            name=name,
            dependencies=dependencies,
            **route_kwargs,
        )(self.get_handler)
        return self



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import asyncio
from pathlib import Path

from typing import Callable, Any, Literal, TYPE_CHECKING, overload, AsyncGenerator, Generator, Generic, TypeVar, cast

if TYPE_CHECKING:
    from .Execution import TriggerFlowExecution
    from .Chunk import TriggerFlowHandler
    from agently.types.data import SerializableValue

from agently.types.trigger_flow import (
    TriggerFlowBlockData,
    TriggerFlowContractMetadata,
    TriggerFlowInterruptEvent,
)
from agently.utils import Settings, RuntimeData, FunctionShifter
from .BluePrint import TriggerFlowBluePrint
from .Process import TriggerFlowProcess
from .Chunk import TriggerFlowChunk
from .Contract import CONTRACT_UNSET, TriggerFlowContract, TriggerFlowContractSpec

InputT = TypeVar("InputT")
StreamT = TypeVar("StreamT")
ResultT = TypeVar("ResultT")
ContractInputT = TypeVar("ContractInputT")
ContractStreamT = TypeVar("ContractStreamT")
ContractResultT = TypeVar("ContractResultT")

class TriggerFlow(Generic[InputT, StreamT, ResultT]):
    def __init__(
        self,
        blue_print: TriggerFlowBluePrint | None = None,
        *,
        name: str | None = None,
        skip_exceptions: bool = False,
    ):
        from agently.base import settings

        self.name = name or uuid.uuid4().hex
        self.settings = Settings(
            name=f"TriggerFlow-{ self.name }-Settings",
            parent=settings,
        )

        self._flow_data = RuntimeData()
        self._runtime_resources = RuntimeData(
            name=f"TriggerFlow-{ self.name }-RuntimeResources",
        )
        self._blue_print = blue_print if blue_print is not None else TriggerFlowBluePrint()
        self._skip_exceptions = skip_exceptions
        self._executions: dict[str, "TriggerFlowExecution[InputT, StreamT, ResultT]"] = {}
        self._contract = TriggerFlowContract[InputT, StreamT, ResultT]()
        self.set_settings = self.settings.set_settings

        self.get_flow_data = self._flow_data.get
        self.set_flow_data = FunctionShifter.syncify(self.async_set_flow_data)
        self.append_flow_data = FunctionShifter.syncify(self.async_append_flow_data)
        self.del_flow_data = FunctionShifter.syncify(self.async_del_flow_data)

        self.start_execution = FunctionShifter.syncify(self.async_start_execution)
        self.register_chunk_handler = self._blue_print.register_chunk_handler
        self.register_condition_handler = self._blue_print.register_condition_handler
        self.set_runtime_resource = self._set_runtime_resource
        self.get_runtime_resource = self._get_runtime_resource
        self.del_runtime_resource = self._del_runtime_resource
        self.update_runtime_resources = self._update_runtime_resources
        self.clear_runtime_resources = self._clear_runtime_resources
        self._bind_start_process()

    def _bind_start_process(self):
        self._start_process = TriggerFlowProcess(
            flow_chunk=self.chunk,
            trigger_event="START",
            blue_print=self._blue_print,
            block_data=TriggerFlowBlockData(
                outer_block=None,
            ),
            definition_signals=[self._blue_print.make_signal("event", "START")],
            definition_group_id=None,
            definition_group_kind=None,
        )
        self.chunks = self._blue_print.chunks
        self.when = self._start_process.when
        self.to = self._start_process.to
        self.to_sub_flow = self._start_process.to_sub_flow
        self.side_branch = self._start_process.side_branch
        self.batch = self._start_process.batch
        self.for_each = self._start_process.for_each
        self.match = self._start_process.match
        self.if_condition = self._start_process.if_condition

    @overload
    def chunk(self, handler_or_name: "TriggerFlowHandler") -> TriggerFlowChunk: ...

    @overload
    def chunk(self, handler_or_name: str) -> "Callable[[TriggerFlowHandler], TriggerFlowChunk]": ...

    def chunk(
        self, handler_or_name: "TriggerFlowHandler | str"
    ) -> "TriggerFlowChunk | Callable[[TriggerFlowHandler], TriggerFlowChunk]":
        if isinstance(handler_or_name, str):

            def wrapper(func: "TriggerFlowHandler"):
                chunk = self._blue_print.create_chunk(
                    func,
                    name=handler_or_name,
                    explicit_name=handler_or_name,
                )
                return chunk

            return wrapper
        else:
            chunk = self._blue_print.create_chunk(
                handler_or_name,
                name=handler_or_name.__name__,
            )
            return chunk

    def create_execution(
        self,
        *,
        skip_exceptions: bool | None = None,
        concurrency: int | None = None,
        runtime_resources: dict[str, Any] | None = None,
    ) -> "TriggerFlowExecution[InputT, StreamT, ResultT]":
        execution_id = uuid.uuid4().hex
        skip_exceptions = skip_exceptions if skip_exceptions is not None else self._skip_exceptions
        execution = self._blue_print.create_execution(
            self,
            execution_id=execution_id,
            skip_exceptions=skip_exceptions,
            concurrency=concurrency,
        )
        if runtime_resources:
            execution.update_runtime_resources(runtime_resources)
        self._executions[execution_id] = execution
        return cast("TriggerFlowExecution[InputT, StreamT, ResultT]", execution)

    @overload
    def set_contract(
        self,
        *,
        meta: dict[str, Any] | None = None,
    ) -> "TriggerFlow[InputT, StreamT, ResultT]": ...

    @overload
    def set_contract(
        self,
        *,
        initial_input: type[ContractInputT],
        meta: dict[str, Any] | None = None,
    ) -> "TriggerFlow[ContractInputT, StreamT, ResultT]": ...

    @overload
    def set_contract(
        self,
        *,
        stream: type[ContractStreamT],
        meta: dict[str, Any] | None = None,
    ) -> "TriggerFlow[InputT, ContractStreamT, ResultT]": ...

    @overload
    def set_contract(
        self,
        *,
        result: type[ContractResultT],
        meta: dict[str, Any] | None = None,
    ) -> "TriggerFlow[InputT, StreamT, ContractResultT]": ...

    @overload
    def set_contract(
        self,
        *,
        initial_input: type[ContractInputT],
        stream: type[ContractStreamT],
        meta: dict[str, Any] | None = None,
    ) -> "TriggerFlow[ContractInputT, ContractStreamT, ResultT]": ...

    @overload
    def set_contract(
        self,
        *,
        initial_input: type[ContractInputT],
        result: type[ContractResultT],
        meta: dict[str, Any] | None = None,
    ) -> "TriggerFlow[ContractInputT, StreamT, ContractResultT]": ...

    @overload
    def set_contract(
        self,
        *,
        stream: type[ContractStreamT],
        result: type[ContractResultT],
        meta: dict[str, Any] | None = None,
    ) -> "TriggerFlow[InputT, ContractStreamT, ContractResultT]": ...

    @overload
    def set_contract(
        self,
        *,
        initial_input: type[ContractInputT],
        stream: type[ContractStreamT],
        result: type[ContractResultT],
        meta: dict[str, Any] | None = None,
    ) -> "TriggerFlow[ContractInputT, ContractStreamT, ContractResultT]": ...

    def set_contract(
        self,
        *,
        initial_input: Any = CONTRACT_UNSET,
        stream: Any = CONTRACT_UNSET,
        result: Any = CONTRACT_UNSET,
        meta: dict[str, Any] | None | object = CONTRACT_UNSET,
    ) -> "TriggerFlow[Any, Any, Any]":
        self._contract.update(
            initial_input=initial_input,
            stream=stream,
            result=result,
            meta=meta,
        )
        self._blue_print.definition.contract = self._contract.export_metadata()
        return self

    def get_contract(self) -> TriggerFlowContractSpec[InputT, StreamT, ResultT]:
        return self._contract.snapshot()

    def get_contract_metadata(self) -> TriggerFlowContractMetadata:
        return self._contract.export_metadata()

    def _set_runtime_resource(self, key: str, value: Any):
        self._runtime_resources.set(str(key), value)
        return self

    def _get_runtime_resource(self, key: str, default: Any = None):
        return self._runtime_resources.get(str(key), default, inherit=False)

    def _del_runtime_resource(self, key: str):
        self._runtime_resources.pop(str(key), None)
        return self

    def _update_runtime_resources(
        self,
        mapping: dict[str, Any] | None = None,
        **kwargs,
    ):
        if mapping is not None:
            for key, value in dict(mapping).items():
                self._set_runtime_resource(str(key), value)
        for key, value in kwargs.items():
            self._set_runtime_resource(str(key), value)
        return self

    def _clear_runtime_resources(self):
        self._runtime_resources.clear()
        return self

    def remove_execution(self, execution: "TriggerFlowExecution | str"):
        if isinstance(execution, str):
            if execution in self._executions:
                del self._executions[execution]
        else:
            if execution.id in self._executions:
                del self._executions[execution.id]

    async def async_start_execution(
        self,
        initial_value: InputT | None,
        *,
        wait_for_result: bool = False,
        concurrency: int | None = None,
        runtime_resources: dict[str, Any] | None = None,
    ) -> "TriggerFlowExecution[InputT, StreamT, ResultT]":
        execution = self.create_execution(
            concurrency=concurrency,
            runtime_resources=runtime_resources,
        )
        await execution.async_start(initial_value, wait_for_result=wait_for_result)
        return execution

    async def _async_change_flow_data(
        self,
        operation: Literal["set", "append", "del"],
        key: str,
        value: Any,
        *,
        emit: bool = True,
    ):
        futures = []
        match operation:
            case "set":
                self._flow_data.set(key, value)
                value = self._flow_data[key]
            case "append":
                self._flow_data.append(key, value)
                value = self._flow_data[key]
            case "del":
                if self._flow_data.get(key, None):
                    del self._flow_data[key]
                    value = None
                else:
                    return

        if emit:
            for execution in self._executions.values():
                handlers = execution._handlers["flow_data"]
                if key in handlers:
                    futures.append(
                        execution.async_emit(
                            key,
                            value,
                            trigger_type="flow_data",
                        )
                    )
            if futures:
                await asyncio.gather(*futures, return_exceptions=True)

    async def async_set_flow_data(
        self,
        key: str,
        value: Any,
        *,
        emit: bool = True,
    ):
        return await self._async_change_flow_data("set", key, value, emit=emit)

    async def async_append_flow_data(
        self,
        key: str,
        value: Any,
        *,
        emit: bool = True,
    ):
        return await self._async_change_flow_data("append", key, value, emit=emit)

    async def async_del_flow_data(
        self,
        key: str,
        *,
        emit: bool = True,
    ):
        return await self._async_change_flow_data("del", key, None, emit=emit)

    @overload
    def start(
        self,
        initial_value: InputT | None = None,
        *,
        wait_for_result: Literal[True] = True,
        timeout: float | None = 10.0,
        concurrency: int | None = None,
        runtime_resources: dict[str, Any] | None = None,
    ) -> ResultT: ...

    @overload
    def start(
        self,
        initial_value: InputT | None = None,
        *,
        wait_for_result: Literal[False],
        timeout: float | None = 10.0,
        concurrency: int | None = None,
        runtime_resources: dict[str, Any] | None = None,
    ) -> None: ...

    def start(
        self,
        initial_value: InputT | None = None,
        *,
        wait_for_result: bool = True,
        timeout: float | None = 10.0,
        concurrency: int | None = None,
        runtime_resources: dict[str, Any] | None = None,
    ) -> ResultT | None:
        return FunctionShifter.syncify(self.async_start)(
            initial_value,
            wait_for_result=wait_for_result,
            timeout=timeout,
            concurrency=concurrency,
            runtime_resources=runtime_resources,
        )

    @overload
    async def async_start(
        self,
        initial_value: InputT | None = None,
        *,
        wait_for_result: Literal[True] = True,
        timeout: float | None = 10.0,
        concurrency: int | None = None,
        runtime_resources: dict[str, Any] | None = None,
    ) -> ResultT: ...

    @overload
    async def async_start(
        self,
        initial_value: InputT | None = None,
        *,
        wait_for_result: Literal[False],
        timeout: float | None = 10.0,
        concurrency: int | None = None,
        runtime_resources: dict[str, Any] | None = None,
    ) -> None: ...

    async def async_start(
        self,
        initial_value: InputT | None = None,
        *,
        wait_for_result: bool = True,
        timeout: float | None = 10.0,
        concurrency: int | None = None,
        runtime_resources: dict[str, Any] | None = None,
    ) -> ResultT | None:
        execution = await self.async_start_execution(
            initial_value,
            concurrency=concurrency,
            runtime_resources=runtime_resources,
        )
        if wait_for_result:
            return await execution.async_get_result(timeout=timeout)

    def get_async_runtime_stream(
        self,
        initial_value: InputT | None = None,
        *,
        timeout: float | None = 10.0,
        concurrency: int | None = None,
        runtime_resources: dict[str, Any] | None = None,
    ) -> AsyncGenerator[StreamT | TriggerFlowInterruptEvent, None]:
        execution = self.create_execution(
            concurrency=concurrency,
            runtime_resources=runtime_resources,
        )
        return execution.get_async_runtime_stream(
            initial_value,
            timeout=timeout,
        )

    def get_runtime_stream(
        self,
        initial_value: InputT | None = None,
        *,
        timeout: float | None = 10.0,
        concurrency: int | None = None,
        runtime_resources: dict[str, Any] | None = None,
    ) -> Generator[StreamT | TriggerFlowInterruptEvent, None, None]:
        execution = self.create_execution(
            concurrency=concurrency,
            runtime_resources=runtime_resources,
        )
        return execution.get_runtime_stream(
            initial_value,
            timeout=timeout,
        )

    def save_blue_print(self):
        return self._blue_print.copy()

    def load_blue_print(self, new_blue_print: TriggerFlowBluePrint):
        self._blue_print = new_blue_print
        self._contract = TriggerFlowContract[InputT, StreamT, ResultT]()
        self.register_chunk_handler = self._blue_print.register_chunk_handler
        self.register_condition_handler = self._blue_print.register_condition_handler
        self._bind_start_process()
        return self

    def get_flow_config(self):
        return self._blue_print.get_flow_config(name=self.name)

    def get_json_flow(
        self,
        save_to: str | Path | None = None,
        *,
        encoding: str | None = "utf-8",
    ):
        return self._blue_print.get_json_flow(
            save_to=save_to,
            encoding=encoding,
            name=self.name,
        )

    def get_yaml_flow(
        self,
        save_to: str | Path | None = None,
        *,
        encoding: str | None = "utf-8",
    ):
        return self._blue_print.get_yaml_flow(
            save_to=save_to,
            encoding=encoding,
            name=self.name,
        )

    def load_flow_config(
        self,
        config: dict[str, Any],
        *,
        replace: bool = True,
    ):
        self._blue_print.load_flow_config(config, replace=replace)
        self._contract = TriggerFlowContract[InputT, StreamT, ResultT]()
        self.name = self._blue_print.name
        self._bind_start_process()
        return self

    def load_json_flow(
        self,
        path_or_content: str | Path,
        *,
        replace: bool = True,
        encoding: str | None = "utf-8",
    ):
        self._blue_print.load_json_flow(
            path_or_content,
            replace=replace,
            encoding=encoding,
        )
        self._contract = TriggerFlowContract[InputT, StreamT, ResultT]()
        self.name = self._blue_print.name
        self._bind_start_process()
        return self

    def load_yaml_flow(
        self,
        path_or_content: str | Path,
        *,
        replace: bool = True,
        encoding: str | None = "utf-8",
    ):
        self._blue_print.load_yaml_flow(
            path_or_content,
            replace=replace,
            encoding=encoding,
        )
        self._contract = TriggerFlowContract[InputT, StreamT, ResultT]()
        self.name = self._blue_print.name
        self._bind_start_process()
        return self

    def to_mermaid(self, *, mode: Literal["simplified", "detailed"] = "simplified"):
        return self._blue_print.to_mermaid(mode=mode, name=self.name)



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import Any, Generic, TypeVar

from pydantic import TypeAdapter, ValidationError

from agently.types.trigger_flow import (
    TRIGGER_FLOW_INTERRUPT_EVENT_SCHEMA,
    TriggerFlowContractEntry,
    TriggerFlowContractMetadata,
    TriggerFlowContractSpec,
    TriggerFlowSystemStreamMetadata,
)

InputT = TypeVar("InputT")
StreamT = TypeVar("StreamT")
ResultT = TypeVar("ResultT")

CONTRACT_UNSET = object()

class TriggerFlowContract(Generic[InputT, StreamT, ResultT]):
    def __init__(
        self,
        *,
        initial_input: Any | None = None,
        stream: Any | None = None,
        result: Any | None = None,
        meta: dict[str, Any] | None = None,
    ):
        self.initial_input: Any | None = None
        self.stream: Any | None = None
        self.result: Any | None = None
        self.meta: dict[str, Any] = {}
        self._initial_input_adapter: TypeAdapter[Any] | None = None
        self._stream_adapter: TypeAdapter[Any] | None = None
        self._result_adapter: TypeAdapter[Any] | None = None
        self.update(
            initial_input=initial_input,
            stream=stream,
            result=result,
            meta=meta,
        )

    def _build_adapter(self, contract: Any | None, contract_name: str):
        if contract is None:
            return None
        try:
            return TypeAdapter(contract)
        except Exception as e:
            raise TypeError(
                f"Invalid TriggerFlow contract '{ contract_name }': { contract }\n"
                "Tips: use a Pydantic model, a Python type, or a supported typing annotation."
            ) from e

    def update(
        self,
        *,
        initial_input: Any = CONTRACT_UNSET,
        stream: Any = CONTRACT_UNSET,
        result: Any = CONTRACT_UNSET,
        meta: dict[str, Any] | None | object = CONTRACT_UNSET,
    ):
        if initial_input is not CONTRACT_UNSET:
            self.initial_input = initial_input
            self._initial_input_adapter = self._build_adapter(initial_input, "initial_input")
        if stream is not CONTRACT_UNSET:
            self.stream = stream
            self._stream_adapter = self._build_adapter(stream, "stream")
        if result is not CONTRACT_UNSET:
            self.result = result
            self._result_adapter = self._build_adapter(result, "result")
        if meta is not CONTRACT_UNSET:
            self.meta = dict(meta) if isinstance(meta, dict) else {}
        return self

    def snapshot(self) -> TriggerFlowContractSpec[InputT, StreamT, ResultT]:
        return TriggerFlowContractSpec[InputT, StreamT, ResultT](
            initial_input=self.initial_input,
            stream=self.stream,
            result=self.result,
            meta=self.meta.copy(),
        )

    def _export_entry(
        self,
        contract: Any | None,
        adapter: TypeAdapter[Any] | None,
    ) -> TriggerFlowContractEntry | None:
        if contract is None:
            return None
        schema = None
        if adapter is not None:
            try:
                schema = adapter.json_schema()
            except Exception:
                schema = None
        label = None
        if isinstance(schema, dict) and isinstance(schema.get("title"), str):
            label = schema["title"]
        if label is None and isinstance(contract, type) and hasattr(contract, "__name__"):
            label = contract.__name__
        if label is None:
            label = str(contract)
        return {
            "label": str(label),
            "schema": schema,
        }

    def export_metadata(self) -> TriggerFlowContractMetadata:
        initial_input = self._export_entry(self.initial_input, self._initial_input_adapter)
        stream = self._export_entry(self.stream, self._stream_adapter)
        result = self._export_entry(self.result, self._result_adapter)
        meta = self.meta.copy()
        system_stream: TriggerFlowSystemStreamMetadata = {
            "interrupt": {
                "label": "TriggerFlowInterruptEvent",
                "schema": copy.deepcopy(TRIGGER_FLOW_INTERRUPT_EVENT_SCHEMA),
            }
        }
        if initial_input is None and stream is None and result is None and not meta:
            return {}
        return {
            "initial_input": initial_input,
            "stream": stream,
            "result": result,
            "meta": meta,
            "system_stream": system_stream,
        }

    def _validate(self, contract_name: str, value: Any, adapter: TypeAdapter[Any] | None):
        if adapter is None:
            return value
        try:
            return adapter.validate_python(value)
        except ValidationError as e:
            raise ValueError(f"TriggerFlow contract validation failed for '{ contract_name }': { e }") from e

    def validate_initial_input(self, value: Any):
        return self._validate("initial_input", value, self._initial_input_adapter)

    def validate_stream_item(self, value: Any):
        return self._validate("stream", value, self._stream_adapter)

    def validate_result(self, value: Any):
        return self._validate("result", value, self._result_adapter)



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

import uuid
import asyncio
from typing import Sequence, TYPE_CHECKING

if TYPE_CHECKING:
    from agently.types.trigger_flow import TriggerFlowRuntimeData

from .BaseProcess import TriggerFlowBaseProcess
from agently.types.data import EMPTY
from agently.types.trigger_flow import TriggerFlowBlockData
from agently.utils import RuntimeDataNamespace


class TriggerFlowForEachProcess(TriggerFlowBaseProcess):
    def for_each(self, *, concurrency: int | None = None):
        for_each_id = uuid.uuid4().hex
        for_each_block_data = TriggerFlowBlockData(
            outer_block=self._block_data,
            data={
                "for_each_id": for_each_id,
                "definition_outer_group_id": self._definition_group_id,
                "definition_outer_group_kind": self._definition_group_kind,
            },
        )
        send_item_trigger = f"ForEach-{ for_each_id }-Send"

        async def send_items(data: "TriggerFlowRuntimeData"):
            data.layer_in()
            for_each_instance_id = data.layer_mark
            assert for_each_instance_id is not None

            send_tasks = []

            def prepare_item(item):
                data.layer_in()
                item_id = data.layer_mark
                assert item_id is not None
                layer_marks = data._layer_marks.copy()
                data._system_runtime_data.set(f"for_each_results.{ for_each_instance_id }.{ item_id }", EMPTY)
                data.layer_out()
                return item_id, layer_marks, item

            async def emit_item(item, layer_marks):
                if concurrency is None or concurrency <= 0:
                    await data.async_emit(
                        send_item_trigger,
                        item,
                        layer_marks,
                    )
                else:
                    semaphore_key = f"for_each_semaphores.{ for_each_id }"
                    semaphore = data._system_runtime_data.get(semaphore_key, inherit=False)
                    if not isinstance(semaphore, asyncio.Semaphore):
                        semaphore = asyncio.Semaphore(concurrency)
                        data._system_runtime_data.set(semaphore_key, semaphore)
                    async with semaphore:
                        await data.async_emit(
                            send_item_trigger,
                            item,
                            layer_marks,
                        )

            if not isinstance(data.value, str) and isinstance(data.value, Sequence):
                items = list(data.value)
                for item in items:
                    _, layer_marks, item_value = prepare_item(item)
                    send_tasks.append(emit_item(item_value, layer_marks))
                await asyncio.gather(*send_tasks)
            else:
                _, layer_marks, item_value = prepare_item(data.value)
                await emit_item(item_value, layer_marks)

        self._blue_print.add_handler(
            self.trigger_type,
            self.trigger_event,
            send_items,
        )
        self._blue_print.definition.add_operator(
            id=f"for_each-split-{ for_each_id }",
            kind="for_each_split",
            name=f"for_each:{ for_each_id }",
            listen_signals=self._definition_signals,
            emit_signals=[self._event_signal(send_item_trigger, role="continuation")],
            options={"concurrency": concurrency},
            group_id=for_each_id,
            group_kind="for_each",
            parent_group_id=self._definition_group_id,
            parent_group_kind=self._definition_group_kind,
        )

        return self._new(
            trigger_event=send_item_trigger,
            trigger_type="event",
            blue_print=self._blue_print,
            block_data=for_each_block_data,
            definition_signals=[self._event_signal(send_item_trigger)],
            definition_group_id=for_each_id,
            definition_group_kind="for_each",
        )

    def end_for_each(self):
        if "for_each_id" not in self._block_data.data:
            raise NotImplementedError("Cannot use .end_for_each() without .for_each().")

        for_each_id = self._block_data.data["for_each_id"]
        end_for_each_trigger = f"ForEach-{ for_each_id }-End"

        async def collect_results(data: "TriggerFlowRuntimeData"):
            for_each_instance_id = data.upper_layer_mark
            item_id = data.layer_mark
            assert for_each_instance_id is not None and item_id is not None

            for_each_results = RuntimeDataNamespace(data._system_runtime_data, "for_each_results")

            if for_each_instance_id in for_each_results and item_id in for_each_results[for_each_instance_id]:
                for_each_results.set(f"{ for_each_instance_id }.{ item_id }", data.value)

                for value in for_each_results.get(for_each_instance_id, {}).values():
                    if value is EMPTY:
                        return

                data.layer_out()
                data.layer_out()

                await data.async_emit(
                    end_for_each_trigger,
                    list(for_each_results[for_each_instance_id].values()),
                    data._layer_marks.copy(),
                )
                for_each_results.delete(for_each_instance_id)

        self._blue_print.add_handler(
            self.trigger_type,
            self.trigger_event,
            collect_results,
        )
        self._blue_print.definition.add_operator(
            id=f"for_each-collect-{ for_each_id }",
            kind="for_each_collect",
            name=f"for_each_collect:{ for_each_id }",
            listen_signals=self._definition_signals,
            emit_signals=[self._event_signal(end_for_each_trigger, role="continuation")],
            group_id=for_each_id,
            group_kind="for_each",
            parent_group_id=self._block_data.data.get("definition_outer_group_id"),
            parent_group_kind=self._block_data.data.get("definition_outer_group_kind"),
        )

        outer_block = self._block_data.outer_block
        block_data = outer_block if outer_block is not None else TriggerFlowBlockData()

        return self._new(
            trigger_event=end_for_each_trigger,
            trigger_type="event",
            blue_print=self._blue_print,
            block_data=block_data,
            definition_signals=[self._event_signal(end_for_each_trigger)],
            definition_group_id=self._block_data.data.get("definition_outer_group_id"),
            definition_group_kind=self._block_data.data.get("definition_outer_group_kind"),
        )
